#define BOOST_TEST_MODULE lock-free-queue
#include <boost/test/included/unit_test.hpp>

#include "lock_free_queue.hpp"
#include <thread>
#include <chrono>
#include <vector>
#include <algorithm>

BOOST_AUTO_TEST_CASE(queue_one_capacity)
{
    add::LockFreeQueue<int, 1> q;
    BOOST_TEST(q.empty());
    BOOST_TEST(!q.full());
    BOOST_TEST(q.size() == 0);
    BOOST_TEST(q.capacity() == 1);
    // push an element
    BOOST_TEST(q.push(1));
    BOOST_TEST(!q.empty());
    BOOST_TEST(q.full());
    BOOST_TEST(q.size() == 1);
    BOOST_TEST(q.capacity() == 1);
    BOOST_TEST(!q.push(2));
    BOOST_TEST(!q.push(3));
    // pop an element
    int val;
    BOOST_TEST((q.pop(val) && val == 1));
    BOOST_TEST(q.empty());
    BOOST_TEST(!q.full());
    BOOST_TEST(q.size() == 0);
    BOOST_TEST(q.capacity() == 1);
}

BOOST_AUTO_TEST_CASE(queue_full)
{
    add::LockFreeQueue<int, 3> q;
    BOOST_TEST(q.push(1));
    BOOST_TEST(q.push(2));
    BOOST_TEST(q.push(3));
    BOOST_TEST(!q.push(4));
}

BOOST_AUTO_TEST_CASE(queue_empty)
{
    add::LockFreeQueue<int, 3> q;
    int val = 0;
    BOOST_TEST(!q.pop(val));
    BOOST_TEST(q.push(1));
    BOOST_TEST(q.push(2));
    BOOST_TEST((q.pop(val) && val == 1));
    BOOST_TEST((q.pop(val) && val == 2));
    BOOST_TEST(!q.pop(val));
}

BOOST_AUTO_TEST_CASE(queue_period)
{
    add::LockFreeQueue<int, 3> q;
    int val = 0;
    BOOST_TEST(q.empty());
    BOOST_TEST(!q.full());
    BOOST_TEST(q.push(1));
    BOOST_TEST(!q.empty());
    BOOST_TEST(!q.full());
    BOOST_TEST(q.push(2));
    BOOST_TEST(q.size() == 2);
    BOOST_TEST(q.capacity() == 3);
    BOOST_TEST(q.push(3));
    BOOST_TEST(q.size() == 3);
    BOOST_TEST(q.capacity() == 3);
    BOOST_TEST(!q.empty());
    BOOST_TEST(q.full());

    // push to full queue
    BOOST_TEST(!q.push(4));
    // pop from full quequ
    BOOST_TEST((q.pop(val) && val == 1));
    BOOST_TEST((q.pop(val) && val == 2));
    // push val
    BOOST_TEST(q.push(4));
    BOOST_TEST(!q.full());
    BOOST_TEST(q.size() == 2);
    BOOST_TEST(q.push(5));
    BOOST_TEST(q.full());
    BOOST_TEST(q.size() == 3);
    BOOST_TEST(!q.push(6));
    BOOST_TEST(q.size() == 3);
    // pop val
    BOOST_TEST((q.pop(val) && val == 3));
    BOOST_TEST((q.pop(val) && val == 4));
    BOOST_TEST((q.pop(val) && val == 5));
    BOOST_TEST(!q.pop(val));
    BOOST_TEST(q.size() == 0);
    BOOST_TEST(!q.full());
    BOOST_TEST(q.empty());
}

BOOST_AUTO_TEST_CASE(queue_period_cap4)
{
    add::LockFreeQueue<int, 4> q;
    int val = 0;
    BOOST_TEST(q.empty());
    BOOST_TEST(!q.full());
    BOOST_TEST(q.push(1));
    BOOST_TEST(!q.empty());
    BOOST_TEST(!q.full());
    BOOST_TEST(q.push(2));
    BOOST_TEST(q.size() == 2);
    BOOST_TEST(q.capacity() == 4);
    BOOST_TEST(q.push(3));
    BOOST_TEST(q.size() == 3);
    BOOST_TEST(q.capacity() == 4);
    BOOST_TEST(q.push(4));
    BOOST_TEST(q.size() == 4);
    BOOST_TEST(q.capacity() == 4);
    BOOST_TEST(!q.empty());
    BOOST_TEST(q.full());

    // push to full queue
    BOOST_TEST(!q.push(4));
    // pop from full quequ
    BOOST_TEST((q.pop(val) && val == 1));
    BOOST_TEST((q.pop(val) && val == 2));
    // push val
    BOOST_TEST(q.push(5));
    BOOST_TEST(!q.full());
    BOOST_TEST(q.size() == 3);
    BOOST_TEST(q.push(6));
    BOOST_TEST(q.full());
    BOOST_TEST(q.size() == 4);
    BOOST_TEST(!q.push(7));
    BOOST_TEST(q.size() == 4);
    // pop val
    BOOST_TEST((q.pop(val) && val == 3));
    BOOST_TEST((q.pop(val) && val == 4));
    BOOST_TEST((q.pop(val) && val == 5));
    BOOST_TEST((q.pop(val) && val == 6));
    BOOST_TEST(!q.pop(val));
    BOOST_TEST(q.size() == 0);
    BOOST_TEST(!q.full());
    BOOST_TEST(q.empty());
}


BOOST_AUTO_TEST_CASE(queue_one_capacity_multithread)
{
    add::LockFreeQueue<int, 1> q;
    volatile bool flag = false;
    auto product_func = [&q, &flag](int begin) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = begin;
        int end = begin + 1000;
        while (i < end) {
            int e = i * 2;
            if (!q.push(e)) {
                continue;
            }
            i++;
        }
    };
    auto consume_func = [&q, &flag](std::vector<int>& backup) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = 0;
        while (i < 1000) {
            int e;
            if (!q.pop(e)) {
                continue;
            }
            backup[i] = e;
            i++;
        }

    };
    std::thread p1(product_func, 0);
    std::thread p2(product_func, 1000);
    std::thread p3(product_func, 2000);
    std::vector<int> backup1(1000);
    std::thread c1 (consume_func, std::ref(backup1));
    std::vector<int> backup2(1000);
    std::thread c2 (consume_func, std::ref(backup2));
    std::vector<int> backup3(1000);
    std::thread c3 (consume_func, std::ref(backup3));
    flag = true;
    p1.join();
    p2.join();
    p3.join();
    c1.join();
    c2.join();
    c3.join();

    std::sort(backup1.begin(), backup1.end());
    std::sort(backup2.begin(), backup2.end());
    std::sort(backup3.begin(), backup3.end());
    BOOST_TEST(q.empty());
    BOOST_TEST(backup1.size() == 1000);
    BOOST_TEST(backup2.size() == 1000);
    BOOST_TEST(backup3.size() == 1000);
    std::vector<int> out1(2000);
    std::vector<int> out(3000);
    std::merge(backup1.begin(), backup1.end(), backup2.begin(), backup2.end(), out1.begin());
    std::merge(out1.begin(), out1.end(), backup3.begin(), backup3.end(), out.begin());
    bool match = true;
    for (int i = 0; i < 3000; i++) {
        if (2 * i != out[i]) {
            match = false;
            break;
        }
    }
    BOOST_TEST(match);
}

struct ElemType {
    int idx;
    int val;
};
BOOST_AUTO_TEST_CASE(queue_single_product)
{
    add::LockFreeQueue<ElemType, 2000> q;
    volatile bool flag = false;
    std::thread p1 ([&q, &flag]() {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = 0;
        while (i < 3000) {
            ElemType e = { i, 2*i };
            if (!q.push(e)) {
                continue;
            }
            i++;
        }
    });
    auto consume_func = [&q, &flag](std::vector<ElemType>& backup) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = 0;
        while (i < 1000) {
            ElemType e;
            if (!q.pop(e)) {
                continue;
            }
            backup[i] = e;
            i++;
        }

    };
    std::vector<ElemType> backup1(1000);
    std::thread c1 (consume_func, std::ref(backup1));
    std::vector<ElemType> backup2(1000);
    std::thread c2 (consume_func, std::ref(backup2));
    std::vector<ElemType> backup3(1000);
    std::thread c3 (consume_func, std::ref(backup3));
    flag = true;
    p1.join();
    c1.join();
    c2.join();
    c3.join();

    auto sortElem = [](const ElemType& v1, const ElemType& v2) { return v1.idx < v2.idx; };
    std::sort(backup1.begin(), backup1.end(), sortElem);
    std::sort(backup2.begin(), backup2.end(), sortElem);
    std::sort(backup3.begin(), backup3.end(), sortElem);
    BOOST_TEST(q.empty());
    BOOST_TEST(backup1.size() == 1000);
    BOOST_TEST(backup2.size() == 1000);
    BOOST_TEST(backup3.size() == 1000);
    std::vector<ElemType> out1(2000);
    std::vector<ElemType> out(3000);
    std::merge(backup1.begin(), backup1.end(), backup2.begin(), backup2.end(), out1.begin(), sortElem);
    std::merge(out1.begin(), out1.end(), backup3.begin(), backup3.end(), out.begin(), sortElem);
    bool match = true;
    for (int i = 0; i < 3000; i++) {
        if (i != out[i].idx || 2 * i != out[i].val) {
            match = false;
            break;
        }
    }
    BOOST_TEST(match);
}

BOOST_AUTO_TEST_CASE(queue_multi_product)
{
    add::LockFreeQueue<ElemType, 2000> q;
    volatile bool flag = false;
    auto product_func = [&q, &flag](int begin) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = begin;
        int end = begin + 1000;
        while (i < end) {
            ElemType e = { i, 2*i };
            if (!q.push(e)) {
                continue;
            }
            i++;
        }
    };
    auto consume_func = [&q, &flag](std::vector<ElemType>& backup) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = 0;
        while (i < 1000) {
            ElemType e;
            if (!q.pop(e)) {
                continue;
            }
            backup[i] = e;
            i++;
        }

    };
    std::thread p1(product_func, 0);
    std::thread p2(product_func, 1000);
    std::thread p3(product_func, 2000);
    std::vector<ElemType> backup1(1000);
    std::thread c1 (consume_func, std::ref(backup1));
    std::vector<ElemType> backup2(1000);
    std::thread c2 (consume_func, std::ref(backup2));
    std::vector<ElemType> backup3(1000);
    std::thread c3 (consume_func, std::ref(backup3));
    flag = true;
    p1.join();
    p2.join();
    p3.join();
    c1.join();
    c2.join();
    c3.join();

    auto sortElem = [](const ElemType& v1, const ElemType& v2) { return v1.idx < v2.idx; };
    std::sort(backup1.begin(), backup1.end(), sortElem);
    std::sort(backup2.begin(), backup2.end(), sortElem);
    std::sort(backup3.begin(), backup3.end(), sortElem);
    BOOST_TEST(q.empty());
    BOOST_TEST(backup1.size() == 1000);
    BOOST_TEST(backup2.size() == 1000);
    BOOST_TEST(backup3.size() == 1000);
    std::vector<ElemType> out1(2000);
    std::vector<ElemType> out(3000);
    std::merge(backup1.begin(), backup1.end(), backup2.begin(), backup2.end(), out1.begin(), sortElem);
    std::merge(out1.begin(), out1.end(), backup3.begin(), backup3.end(), out.begin(), sortElem);
    bool match = true;
    for (int i = 0; i < 3000; i++) {
        if (i != out[i].idx || 2 * i != out[i].val) {
            match = false;
            break;
        }
    }
    BOOST_TEST(match);
}

class EntryType {
public:
    EntryType() {}
    EntryType(int idx) : idx_(idx) {}
    EntryType(const EntryType& other) : idx_(other.idx_) { move_ = other.move_; }
    EntryType(EntryType&& other) : idx_(std::move(other.idx_)) {
        move_ = true;
    }
    EntryType& operator=(EntryType&& other) {
        if (&other == this) {
            return *this;
        }
        idx_ = std::move(other.idx_);
        move_ = true;
        return *this;
    }
    EntryType& operator=(const EntryType& other) {
        if (&other == this) {
            return *this;
        }
        idx_ = other.idx_;
        move_ = other.move_;
        return *this;
    }
    int GetId() const { return idx_; }
    bool GetMoveFlag() const { return move_; }
private:
    int idx_ { 0 };
    bool move_ { false };
};
BOOST_AUTO_TEST_CASE(queue_push_move)
{
    add::LockFreeQueue<EntryType, 2000> q;
    volatile bool flag = false;
    auto product_func = [&q, &flag](int begin) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = begin;
        int end = begin + 1000;
        while (i < end) {
            EntryType e(i);
            if (!q.push(std::move(e))) {
                continue;
            }
            i++;
        }
    };
    auto consume_func = [&q, &flag](std::vector<EntryType>& backup) {
        while (!flag) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(1));
        }
        int i = 0;
        while (i < 1000) {
            EntryType e;
            if (!q.pop(e)) {
                continue;
            }
            backup[i] = e;
            i++;
        }

    };
    std::thread p1(product_func, 0);
    std::thread p2(product_func, 1000);
    std::thread p3(product_func, 2000);
    std::vector<EntryType> backup1(1000);
    std::thread c1 (consume_func, std::ref(backup1));
    std::vector<EntryType> backup2(1000);
    std::thread c2 (consume_func, std::ref(backup2));
    std::vector<EntryType> backup3(1000);
    std::thread c3 (consume_func, std::ref(backup3));
    flag = true;
    p1.join();
    p2.join();
    p3.join();
    c1.join();
    c2.join();
    c3.join();

    // std::sort may call move operator=, so, chech move flag before std::sort
    bool called_move = true;
    for (int i = 0; i < 1000; i++) {
        if (!backup1[i].GetMoveFlag()) { called_move = false; break; }
        if (!backup2[i].GetMoveFlag()) { called_move = false; break; }
        if (!backup3[i].GetMoveFlag()) { called_move = false; break; }
    }
    BOOST_TEST(called_move);

    auto sortElem = [](const EntryType& v1, const EntryType& v2) { return v1.GetId() < v2.GetId(); };
    std::sort(backup1.begin(), backup1.end(), sortElem);
    std::sort(backup2.begin(), backup2.end(), sortElem);
    std::sort(backup3.begin(), backup3.end(), sortElem);
    BOOST_TEST(q.empty());
    BOOST_TEST(backup1.size() == 1000);
    BOOST_TEST(backup2.size() == 1000);
    BOOST_TEST(backup3.size() == 1000);
    std::vector<EntryType> out1(2000);
    std::vector<EntryType> out(3000);
    std::merge(backup1.begin(), backup1.end(), backup2.begin(), backup2.end(), out1.begin(), sortElem);
    std::merge(out1.begin(), out1.end(), backup3.begin(), backup3.end(), out.begin(), sortElem);
    bool match = true;
    for (int i = 0; i < 3000; i++) {
        if (i != out[i].GetId()) {
            match = false;
            break;
        }
    }
    BOOST_TEST(match);
}

